import pickle
from copy import deepcopy
from dataclasses import asdict, dataclass
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp2d

from nuance.utils import interp_split_times


@dataclass
class SearchData:
    """
    An object that holds the results of the search.
    """

    # linear search grid
    t0s: np.ndarray
    """Array of searched signal trial epochs."""
    Ds: np.ndarray
    """Array of searched signal trial durations."""

    # linear search results
    ll: Optional[np.ndarray] = None
    """Non-periodic signal likelihoods."""
    z: Optional[np.ndarray] = None
    """Non-periodic signal depths."""
    vz: Optional[np.ndarray] = None
    """Non-periodic signal depth variance."""
    ll0: Optional[float] = None
    """No-signal likelihood."""

    # periodic search, Q is periodogram
    periods: Optional[np.ndarray] = None
    """Array of signal trial periods."""
    Q_snr: Optional[np.ndarray] = None
    """Periodogram SNR."""
    Q_ll: Optional[np.ndarray] = None
    """Periodic signal likelihoods."""
    Q_params: Optional[np.ndarray] = None
    """Periodogram best-fit parameters."""

    @property
    def _folds(self):
        f_ll = interp2d(self.Ds, self.t0s, self.ll)
        f_z = interp2d(self.Ds, self.t0s, self.z)
        f_dz2 = interp2d(self.Ds, self.t0s, self.vz)

        def interpolate_all(pt0s):
            # computing the likelihood folds by interpolating in phase
            folds_ll = np.array([f_ll(self.Ds, t) for t in pt0s])
            folds_z = np.array([f_z(self.Ds, t) for t in pt0s])
            folds_vz = np.array([f_dz2(self.Ds, t) for t in pt0s])

            return folds_ll, folds_z, folds_vz

        def _folds(p, dphi=0.01):
            pt0s = interp_split_times(self.t0s, p, dphi=dphi)
            return pt0s, interpolate_all(pt0s)

        return _folds

    @property
    def fold_ll(self):
        """
        Returns a function that folds the signal likelihoods of all periods and t0s into a
        single period-folded likelihood.

        Returns:
            function: A function that takes a period and returns the folded likelihoods.
        """

        folds = self._folds

        def _fold(p, dphi=0.01):
            pt0s, (lls, zs, vzs) = folds(p, dphi=dphi)
            P1 = np.sum(lls, 0)
            vZ = 1 / np.sum(1 / vzs, 0)
            Z = vZ * np.sum(zs / vzs, 0)
            P1 = np.sum(lls, 0)
            P2 = 0.5 * np.sum(
                np.log(vzs) - np.log(vzs + vZ) + (zs - Z) ** 2 / (vzs + vZ), 0
            )

            return pt0s[0] / p, P1, P1 - P2

        return _fold

    @property
    def best(self):
        """
        Returns the best-fit parameters for the search object.

        Returns:
            tuple: A tuple containing the best-fit t0, D, and period (if available).
        """
        assert (
            self.ll is not None and self.Q_params is not None and self.Q_snr is not None
        ), "No search performed"
        if self.periods is not None:
            i = np.argmax(self.Q_snr)
            return self.Q_params[i]
        else:
            i, j = np.unravel_index(np.argmax(self.ll), self.ll.shape)
            t0, D = self.t0s[i], self.Ds[j]
            period = None
        return t0, D, period

    @property
    def shape(self):
        """
        Returns the shape of the likelihood array.

        Returns:
            tuple: A tuple containing the number of t0s and Ds in the search objects.
        """
        return len(self.t0s), len(self.Ds)

    def show_ll(self, **kwargs):
        """
        Plots the likelihood array.

        Args:
            **kwargs: Additional keyword arguments to pass to `plt.imshow`.
        """
        assert self.ll is not None, "No search performed"
        extent = np.min(self.t0s), np.max(self.t0s), np.min(self.Ds), np.max(self.Ds)
        plt.imshow(self.ll.T, aspect="auto", origin="lower", extent=extent, **kwargs)

    def asdict(self):
        """
        Returns a dictionary representation of the search object.

        Returns:
            dict: A dictionary containing the search object parameters.
        """
        return asdict(self)

    def save(self, filename):
        """
        Saves the search to a file.

        Args:
            filename (str): The name of the file to save the search object to.
        """

        pickle.dump(self.asdict(), open(filename, "wb"))

    def copy(self):
        """
        Returns a deep copy of the search object.

        Returns:
            Search: A deep copy of the search object.
        """
        return deepcopy(self)

    @classmethod
    def load(cls, filename):
        """
        Loads a search object from a file.

        Args:
            filename (str): The name of the file to load the search object from.

        Returns:
            Search: The loaded search object.
        """
        return cls(**pickle.load(open(filename, "rb")))



import jax
import matplotlib.pyplot as plt
import numpy as np
import tinygp

from nuance import core
from nuance.core import transit_protopapas


def interp_split_times(time, p, dphi=0.01):
    tmax, tmin = np.max(time), np.min(time)
    total = tmax - tmin
    # since for very small periods we might fold on few points only, it's better to impose
    # at least 200 phases, to compare period folds more fairly
    phase = np.arange(0, 1, np.min([1 / 200, dphi / p]))
    n = np.arange(np.ceil(total / p) + 1)  # number of 'folds'
    # this line is important so that different time-series have the same phase 0
    tmin -= tmin % p
    pt0s = np.array([tmin + phase * p + j * p for j in n])  # corresponding t0s
    has_time = np.any([np.abs(time - _t0) < p / 2 for _t0 in pt0s.mean(1)], 1)
    pt0s = pt0s[has_time]

    return pt0s


def phase(t, t0, p):
    return (t - t0 + 0.5 * p) % p - 0.5 * p


def tv_dv(duration, depth, omega, sigma):
    return np.pi / (omega * duration), 2 * sigma / depth


def binn(x, y, n):
    N = int(len(x) / n)
    ns = np.histogram(x, N)[0]
    bx = np.histogram(x, N, weights=x)[0] / ns
    by = np.histogram(x, N, weights=y)[0] / ns
    return bx, by


def log_gauss_product_integral(a, va, b, vb):
    """Log of the product ot two Normal distributions (normalized)

    Parameters
    ----------
    a : float
        mean of distribution A
    va : float
        variance of distribution A
    b : float
        mean of distribution B
    vb : float
        variance of distribution B

    Returns
    -------
    float
        log of the integral of A*B
    """
    return (
        -0.5 * np.square(a - b) / (va + vb)
        - 0.5 * np.log(va + vb)
        - 0.5 * np.log(np.pi)
        - np.log(2) / 2
    )


def Ps(lls, zs, vzs):
    vZ = 1 / np.sum(1 / vzs, 0)
    Z = vZ * np.sum(zs / vzs, 0)

    P1 = np.sum(lls, 0)
    P2 = P1 + np.sum(log_gauss_product_integral(zs, vzs, Z, vZ), 0)
    return P1, P2


def simulated(
    t0=0.2, D=0.05, depth=0.02, P=0.7, time=None, kernel=None, error=0.001, w=None
):
    if time is None:
        time = np.arange(0, 4, 2 / 60 / 24)

    X = np.vander(time, N=4, increasing=True).T
    if w is None:
        w = [1.0, 5e-4, -2e-4, -5e-4]

    true_transit = depth * core.transit_protopapas(time, t0, D, P=P)

    if kernel is None:
        kernel = tinygp.kernels.quasisep.SHO(np.pi / (6 * D), 45.0, depth)

    gp = tinygp.GaussianProcess(kernel, time, diag=error**2, mean=0.0)
    flux = gp.sample(jax.random.PRNGKey(40)) + true_transit + w @ X

    return (time, flux, error), X, gp


def plot_search(nu, search, bins=7 / 60 / 24):
    """Plot result of a the periodic search

    Parameters
    ----------
    nu : _type_
        _description_
    search : _type_
        _description_
    """
    t0, D, P = search.best

    plt.subplot(2, 2, (1, 3))
    plt.plot(search.periods, search.Q_snr)
    plt.title(f"{P:.5f} days")

    mean, astro, noise = nu.models(t0, D, P)

    plt.subplot(2, 2, 2)
    plt.plot(nu.flux - mean, ".", c="0.8")
    plt.plot(astro, c="k", label="found")
    ylim = plt.ylim()
    _ = plt.legend()

    plt.subplot(2, 2, 4)
    mean, astro, noise = nu.models(t0, D, P)
    phi = phase(nu.time, t0, P)
    detrended = nu.flux - noise - mean
    plt.plot(phi, detrended, ".", c=".8")
    bx, by, be = binn_time(phi, detrended, bins=bins)
    plt.errorbar(bx, by, yerr=be, fmt=".", c="k")
    plt.xlim(*(np.array([-1, 1]) * 10 * D))
    plt.ylim(*(np.array([-1, 1]) * float(np.abs(astro.min())) * 4))


def clean_periods(periods, period, tol=0.02):
    close = periods / period
    return periods[np.abs(close - np.rint(close)) > tol]


def index_binning(x, size):
    if isinstance(size, float):
        bins = np.arange(np.min(x), np.max(x), size)
    else:
        x = np.arange(0, len(x))
        bins = np.arange(0.0, len(x), size)

    d = np.digitize(x, bins)
    n = np.max(d) + 2
    indexes = []

    for i in range(0, n):
        s = np.where(d == i)
        if len(s[0]) > 0:
            s = s[0]
            indexes.append(s)

    return indexes


def binn_time(time, flux, bins=7 / 24 / 60):
    indexes = index_binning(time, bins)
    binned_time = np.array([np.mean(time[i]) for i in indexes])
    binned_flux = np.array([np.mean(flux[i]) for i in indexes])
    binned_error = np.array([np.std(flux[i]) / np.sqrt(len(i)) for i in indexes])
    return binned_time, binned_flux, binned_error


def simulated_ground_based(n=500, N=4):
    true = {
        "t0": 0.3,
        "D": 38 / 60 / 24,
        "depth": 0.005,
        "P": 0.65,
    }
    idxs = np.arange(1, N) * n
    times = [np.linspace(0 + i, 0.5 + i, n) for i in range(N)]
    time = np.concatenate(times)
    error = 1e-3
    kernel = tinygp.kernels.quasisep.SHO(np.pi / true["D"] / 3, 1, true["depth"] / 2)
    variability_gp = tinygp.GaussianProcess(kernel, time, diag=0)
    jax_key = jax.random.PRNGKey(0)
    variability = variability_gp.sample(jax_key)
    variabilities = [
        np.random.normal(0, error) + v for v in np.split(variability, idxs)
    ]

    airmasses = [0.2 * (t - t.min()) ** 2 for t in times]
    bkgs = [
        tinygp.GaussianProcess(
            tinygp.kernels.quasisep.SHO(20, 1, 0.005), t, diag=(1e-4) ** 2
        ).sample(jax.random.PRNGKey(i))
        for i, t in enumerate(times)
    ]
    fwhms = [
        tinygp.GaussianProcess(
            tinygp.kernels.quasisep.SHO(45, 1, 0.005), t, diag=(5e-4) ** 2
        ).sample(jax.random.PRNGKey(i))
        for i, t in enumerate(times)
    ]

    transits = np.split(
        true["depth"] * transit_protopapas(time, 0.3, true["D"], true["P"]), idxs
    )

    systematics = [
        0.2 * np.random.normal(0, 0.9, 3) @ np.vstack([airmasses[i], bkgs[i], fwhms[i]])
        for i in range(N)
    ]

    fluxes = [systematics[i] + variabilities[i] + transits[i] + 1.0 for i in range(N)]

    fluxes = [f - np.median(f) + 1.0 for f in fluxes]

    observations = [
        {
            "time": times[i],
            "flux": fluxes[i],
            "error": np.ones_like(times[i]) * error,
            "airmass": airmasses[i],
            "bkg": bkgs[i],
            "fwhm": fwhms[i],
        }
        for i in range(N)
    ]

    signals = {
        "transits": transits,
        "variabilities": variabilities,
        "systematics": systematics,
        "transit_params": true,
    }

    return observations, signals



import multiprocessing as mp
import pickle
from copy import deepcopy
from dataclasses import asdict, dataclass
from functools import partial

import jax
import jax.numpy as jnp
import jaxopt
import multiprocess as mp
import numpy as np
from scipy.ndimage import minimum_filter1d
from tinygp import GaussianProcess, kernels
from tqdm.auto import tqdm

from nuance import DEVICES_COUNT, core, utils
from nuance.search_data import SearchData


@dataclass
class Nuance:
    """
    `nuance` object containing the data and the linear and periodic search functions.

    .. note::

        `nuance` relies on `tinygp <https://github.com/dfm/tinygp>`_ to instantiate and manipulate Gaussian processes. See the
        `tinygp documentation <https://tinygp.readthedocs.io>`_ for more details.

    Parameters
    ----------
    time : np.ndarray
        Time
    flux : np.ndarray
        Flux time series
    error : np.ndarray, optional
        Flux error time series, by default None
    gp : tinygp GaussianProcess, optional
        `tinygp.GaussianProcess <https://tinygp.readthedocs.io/en/latest/api/summary/tinygp.GaussianProcess.html>`_ instance, by default None,
        which creates a GP with a squared exponential kernel of lengthscale :math:`10^{12}` and sigma of  :math:`1`.
    X : np.ndarray, optional
        Design matrix, by default None, which create a design matrix with a single column of ones.
    compute : bool, optional
        Whether to pre-compute the Cholesky decomposition used by the GP, by default True.
    mean : float, optional
        Mean of the GP, by default 0.0
    search_data : SearchData, optional
        Search data instance, by default None.
    model : callable, optional
        Model function with signature `model(time, t0, D, P=None)`, by default None, which set the model to
        :code:`partial(nuance.core.transit_protopapas, c=12)`
    """

    time: np.ndarray
    """Time"""
    flux: np.ndarray
    """Flux time series"""
    error: np.ndarray = None
    """Flux error time series"""
    gp: GaussianProcess = None
    """tinygp.GaussianProcess instance"""
    X: np.ndarray = None
    """Design matrix"""
    compute: bool = True
    """Whether to pre-compute the Cholesky decomposition"""
    mean: float = 0.0
    """Mean of the GP"""
    search_data: SearchData = None
    """Search data instance"""
    model: callable = None
    """Model function with signature :code:`model(time, t0, D, P=None)`"""

    def __post_init__(self):
        if self.model is None:
            self.model = partial(core.transit_protopapas, c=12)

        self.model = jax.jit(self.model)

        assert (self.error is None) ^ (
            self.gp is None
        ), "Either error or gp must be defined"

        if self.X is None:
            self.X = np.atleast_2d(np.ones_like(self.time))

        if self.gp is None:
            kernel = kernels.quasisep.Exp(1e12)

            self.gp = GaussianProcess(
                kernel, self.time, diag=self.error**2, mean=self.mean
            )

        if self.compute:
            self.eval_model = jax.jit(core.eval_model(self.flux, self.X, self.gp))

        self.search_data = None

    def __repr__(self):
        noise = (
            f"kernel={self.gp.kernel}"
            if self.error is None
            else f"error={self.error:.3e}"
        )
        return f"Nuance(N={len(self.time)}, M={self.X.shape[0]}, {noise}, searched={self.searched})"

    @property
    def searched(self):
        """Whether the linear search has been performed."""
        return self.search_data is not None

    @property
    def time_span(self):
        """Time span"""
        return np.max(self.time) - np.min(self.time)

    @property
    def ll0(self) -> float:
        """Log-likelihood of data without model.

        Returns
        -------
        float
        """
        return self.eval_model(np.zeros_like(self.time))[0].__array__()

    def _models(self, m):
        _, w, _ = self.eval_model(m)
        mean = w[0:-1] @ self.X
        signal = m * w[-1]

        @jax.jit
        def gp_mean():
            return self.gp.condition(self.flux - mean - signal).gp.mean

        noise = gp_mean()

        return mean, signal, noise

    def gp_optimization(self, build_gp, mask=None):
        """
        Optimization functions to fit a Gaussian Process given a building function.

        Parameters
        ----------
        build_gp : function
            A function that returns a tinygp.GaussianProcess object.
        mask : array-like, optional
            A boolean array to mask the data, by default None.

        Returns
        -------
        tuple
            A tuple containing three functions:

            - :code:`optimize`: a function that optimizes the GP model.
            - :code:`mu`: a function that returns the mean of the GP model (jax.jit-compiled).
            - :code:`nll`: a function that returns the negative log-likelihood of the GP model (jax.jit-compiled).
        """
        if mask is None:
            mask = np.ones_like(self.time).astype(bool)

        masked_x = self.time[mask]
        masked_y = self.flux[mask]
        masked_X = self.X[:, mask]

        @jax.jit
        def nll_w(params):
            gp = build_gp(params, masked_x)
            Liy = gp.solver.solve_triangular(masked_y)
            LiX = gp.solver.solve_triangular(masked_X.T)
            LiXT = LiX.T
            LiX2 = LiXT @ LiX
            w = jnp.linalg.lstsq(LiX2, LiXT @ Liy)[0]
            nll = -gp.log_probability(masked_y - w @ masked_X)
            return nll, w

        @jax.jit
        def nll(params):
            return nll_w(params)[0]

        @jax.jit
        def mu(params):
            gp = build_gp(params, masked_x)
            _, w = nll_w(params)
            cond_gp = gp.condition(masked_y - w @ masked_X, self.time).gp
            return cond_gp.loc + w @ self.X

        def optimize(init_params, param_names=None):
            def inner(theta, *args, **kwargs):
                params = dict(init_params, **theta)
                return nll(params, *args, **kwargs)

            param_names = (
                list(init_params.keys()) if param_names is None else param_names
            )
            start = {k: init_params[k] for k in param_names}

            solver = jaxopt.ScipyMinimize(fun=inner)
            soln = solver.run(start)
            print(soln.state)

            return dict(init_params, **soln.params)

        return optimize, mu, nll

    def mu(self, time=None):
        """
        Computes the mean model of the GP.

        Parameters
        ----------
        time : np.ndarray, optional
            The time at which to compute the mean model, by default None (uses `self.time`).
        Returns
        -------
        np.ndarray
            The mean model of the GP.
        """
        if time is None:
            time = self.time

        @jax.jit
        def _mu():
            gp = self.gp
            _, w, _ = self.eval_model(np.zeros_like(time))
            w = w[0:-1]
            cond_gp = gp.condition(self.flux - w @ self.X, time).gp
            return cond_gp.loc + w @ self.X

        return _mu()

    def models(self, t0: float = None, D: float = None, P: float = 1e15):
        """Return the models corresponding to epoch `t0` and duration `D`(and period `P` for a periodic model).

        Parameters
        ----------
        t0 : float
            epoch, same unit as time
        D : float
            duration, same unit as time
        P : float, optional
            period, same unit as time, by default None

        Returns
        -------
        list np.ndarray
            a list of three np.ndarray:

            - linear: linear model (using `X`)
            - model: model being searched
            - noise: noise model (using `GP`)

        Example
        -------

        .. code-block::

            from nuance import Nuance, utils

            time, flux, error = utils.simulation()[0]

            nu = Nuance(time, flux, error)
            linear, astro, noise = nu.models(0.2, 0.05, 1.3)

        """
        if t0 is not None:
            m = self.model(self.time, t0, D, P)
        else:
            m = np.zeros_like(self.time)
        return self._models(m)

    def solve(self, t0: float, D: float, P: float = None):
        """Solve linear model (suing `X`).

        Parameters
        ----------
        t0 : float
            epoch, same unit as time
        D : float
            duration, same unit as time
        P : float, optional
            period, same unit as time, by default None

        Returns
        -------
        list
            (w, v): linear coefficients and their covariance matrix
        """
        m = self.model(self.time, t0, D, P)
        _, w, v = self.eval_model(m)
        return w, v

    def depth(self, t0: float, D: float, P: float = None):
        """Depth linearly solved for epoch `t0` and duration `D` (and period `P` for a periodic model).

        Parameters
        ----------
        t0 : float
            epoch, same unit as time
        D : float
            duration, same unit as time
        P : float, optional
            period, same unit as time, by default None

        Returns
        -------
        float, float
            model depth, depth error
        """
        w, v = self.solve(t0, D, P)
        return w[-1], jnp.sqrt(v[-1, -1])

    def snr(self, t0: float, D: float, P: float = None):
        """SNR of the model linearly solved for epoch `t0` and duration `D` (and period `P` for a periodic model).

        Parameters
        ----------
        t0 : float
            epoch, same unit as time
        D : float
            duration, same unit as time
        P : float, optional
            period, same unit as time, by default None

        Returns
        -------
        float
            model snr
        """
        w, dw = self.depth(t0, D, P=P)
        return jnp.max(jnp.array([0, w / dw]))

    def linear_search(
        self,
        t0s: np.ndarray,
        Ds: np.ndarray,
        positive: bool = True,
        progress: bool = True,
        backend: str = None,
        batch_size: int = None,
    ):
        """Performs the linear search. Saves the linear search `Nuance.search_data` as a :py:class:`nuance.SearchData` object.

        Parameters
        ----------
        t0s : np.ndarray
            array of model epochs
        Ds : np.ndarray
            array of model durations
        positive : bool, optional
            wether to force depth to be positive, by default True
        progress : bool, optional
            wether to show progress bar, by default True
        backend : str, optional
            backend to use, by default jax.default_backend() (options: "cpu", "gpu").
            This affects the linear search function jax-mapping strategy. For more details, see
            :py:func:`nuance.core.map_function`
        batch_size : int, optional
            batch size for parallel evaluation, by default None
        Returns
        -------
        None
        """
        assert backend in [None, "cpu", "gpu"], "backend must be 'cpu' or 'gpu'"

        if backend is None:
            backend = jax.default_backend()

        if backend == "cpu":
            eval_t0_Ds_function = core.pmap_cpus
            if batch_size is None:
                batch_size = DEVICES_COUNT

        elif backend == "gpu":
            eval_t0_Ds_function = core.vmap_gpu
            if batch_size is None:
                batch_size = 1000

        eval_t0s_Ds = eval_t0_Ds_function(self.eval_model, self.model, self.time)

        batches_n = int(np.ceil(len(t0s) / batch_size))
        padded_t0s = np.pad(t0s, pad_width=[0, batches_n * batch_size - len(t0s)])
        batched_t0s = np.array(np.array_split(padded_t0s, batches_n))

        ll = np.zeros((len(padded_t0s), len(Ds)))
        depths = ll.copy()
        vars = ll.copy()
        depths = ll.copy()

        _progress = lambda x: (tqdm(x) if progress else x)

        for i, t0 in enumerate(_progress(batched_t0s)):
            _depths, _vars, _ll = eval_t0s_Ds(t0, Ds)
            depths[i * batch_size : (i + 1) * batch_size, :] = _depths.T
            vars[i * batch_size : (i + 1) * batch_size, :] = _vars.T
            ll[i * batch_size : (i + 1) * batch_size, :] = _ll.T

        depths = np.array(depths[0 : len(t0s), :])
        vars = np.array(vars[0 : len(t0s), :])
        ll = np.array(ll[0 : len(t0s), :])

        if positive:
            ll0 = self.eval_model(np.zeros_like(self.time))[0]
            ll[depths < 0] = ll0

        vars[~np.isfinite(vars)] = 1e25

        self.search_data = SearchData(
            t0s=t0s, Ds=Ds, ll=ll, z=depths, vz=vars, ll0=self.ll0
        )

    def periodic_search(self, periods: np.ndarray, progress=True, dphi=0.01):
        """Performs the periodic search.

        Parameters
        ----------
        periods : np.ndarray
            array of periods to search
        progress : bool, optional
            wether to show progress bar, by default True
        dphi: float, optional
            the relative step size of the phase grid. For each period, all likelihood quantities along time are
            interpolated along a phase grid of resolution `min(1/200, dphi/P))`. The smaller dphi
            the finer the grid, and the more resolved the model epoch and period (but the more computationally expensive the
            periodic search). The default is 0.01.

        Returns
        -------
        :py:class:`~nuance.SearchData`
            search results
        """
        new_search_data = self.search_data.copy()
        n = len(periods)
        snr = np.zeros(n)
        params = np.zeros((n, 3))

        def _progress(x, **kwargs):
            return tqdm(x, **kwargs) if progress else x

        global SEARCH
        SEARCH = partial(self.search_data.fold_ll, dphi=dphi)

        with mp.Pool() as pool:
            for p, (Ti, j, P) in enumerate(
                _progress(pool.imap(_search, periods), total=len(periods))
            ):
                Dj = new_search_data.Ds[j]
                pass
                snr[p], params[p] = float(self.snr(Ti, Dj, P)), (Ti, Dj, P)

        new_search_data.periods = periods
        new_search_data.Q_snr = snr
        new_search_data.Q_params = params

        return new_search_data

    def mask_model(self, t0: float, D: float, P: float):
        """Return a `Solver` object where the model of epoch `t0` and duration `D` and period `P` is masked.

        Parameters
        ----------
        t0 : float
            model epoch
        D : float
            model duration
        P : float
            model period

        Returns
        -------
        `Nuance`
            A `Nuance` instance with model masked.
        """
        # search data
        search_data = self.search_data.copy()
        ph = utils.phase(search_data.t0s, t0, P)
        t0_mask = np.abs(ph) > 2 * D

        # if np.count_nonzero(t0_mask) == len(search_data.t0s):
        #     raise ValueError("Mask covers all data points")
        # elif len(t0_mask) == 0:
        #     raise ValueError("No data to mask")

        search_data.llv = None
        search_data.llc = None
        search_data.periods = None
        search_data.t0s = search_data.t0s[t0_mask]
        search_data.ll = search_data.ll[t0_mask]
        search_data.z = search_data.z[t0_mask]
        search_data.vz = search_data.vz[t0_mask]

        # nu
        ph = utils.phase(self.time, t0, P)
        t0_mask = np.abs(ph) > 2 * D

        ph = utils.phase(self.time, t0, P)
        time_mask = np.abs(ph) > 2 * D
        gp = GaussianProcess(
            self.gp.kernel,
            self.time[time_mask],
            mean=0.0,
            diag=self.gp.variance[time_mask],
        )

        nu = Nuance(
            self.time[time_mask],
            self.flux[time_mask],
            gp=gp,
            X=self.X[:, time_mask],
            mean=0.0,
        )
        nu.search_data = search_data

        return nu

    def save(self, filename):
        """Save the current state of the object to a file.

        Parameters
        ----------
        filename : str
            The name of the file to save the object to.
        """
        pickle.dump(asdict(self), open(filename, "wb"))

    def copy(self):
        """Return a deep copy of the object."""
        return deepcopy(self)

    @classmethod
    def load(cls, filename):
        """Load an object from a file.

        Parameters
        ----------
        filename : str
            The name of the file to load the object from.

        Returns
        -------
        Nuance
            The loaded object.
        """
        return cls(**pickle.load(open(filename, "rb")))

    def mask_flares(self, build_gp=None, init=None, window=20, sigma=5, iterations=3):
        # for now
        assert build_gp is not None and init is not None
        mask = np.ones_like(self.time).astype(bool)

        if build_gp is not None:
            optimize, mu, nll = self.gp_optimization(build_gp)
            opt = init.copy()

        for _ in range(iterations):
            residuals = self.flux - mu(opt)
            mask[residuals > sigma * np.std(residuals[mask])] = False
            mask = np.roll(minimum_filter1d(mask, window), shift=window // 3)

            if build_gp is not None:
                optimize, mu, _ = self.gp_optimization(build_gp, mask)
                opt = optimize(init)

        new_nu = Nuance(
            time=self.time[mask],
            flux=self.flux[mask],
            X=self.X[:, mask],
            gp=build_gp(opt, self.time[mask]),
            compute=True,
        )

        return new_nu


def _search(p):
    phase, _, P2 = SEARCH(p)
    i, j = np.unravel_index(np.argmax(P2), P2.shape)
    Ti = phase[i] * p
    return Ti, j, p



from functools import partial

import jax
import jax.numpy as jnp


def eval_model(flux, X, gp):
    Liy = gp.solver.solve_triangular(flux)
    LiX = gp.solver.solve_triangular(X.T)

    @jax.jit
    def function(m):
        Xm = jnp.vstack([X, m])
        Lim = gp.solver.solve_triangular(m)
        LiXm = jnp.hstack([LiX, Lim[:, None]])
        LiXmT = LiXm.T
        LimX2 = LiXmT @ LiXm
        w = jnp.linalg.lstsq(LimX2, LiXmT @ Liy)[0]
        v = jnp.linalg.inv(LimX2)
        return gp.log_probability(flux - w @ Xm), w, v

    return function


def transit_protopapas(t, t0, D, P=1e15, c=12):
    _t = P * jnp.sin(jnp.pi * (t - t0) / P) / (jnp.pi * D)
    return -0.5 * jnp.tanh(c * (_t + 1 / 2)) + 0.5 * jnp.tanh(c * (_t - 1 / 2))


def transit_box(time, t0, D, P=1e15):
    return -((jnp.abs(time - t0) % P) < D / 2).astype(float)


def transit_exocomet(time, t0, duration, P=None, n=3):
    flat = jnp.zeros_like(time)
    left = -(time - (t0 - duration / n)) / (duration / n)
    right = -jnp.exp(-2 / duration * (time - t0 - duration / n)) ** 2
    triangle = jnp.maximum(left, right)
    mask = time >= t0 - duration / n
    signal = jnp.where(mask, triangle, flat)
    return signal / jnp.max(jnp.array([-jnp.min(signal), 1]))


def map_function(eval_function, model, time, backend, map_t0, map_D):
    jitted_eval = jax.jit(eval_function, backend=backend)

    @jax.jit
    def single_eval(t0, D):
        m = model(time, t0, D)
        ll, w, v = jitted_eval(m)
        return w[-1], v[-1, -1], ll

    t0s_eval = map_t0(single_eval, in_axes=(0, None))
    ds_t0s_eval = map_D(t0s_eval, in_axes=(None, 0))

    return ds_t0s_eval


pmap_cpus = partial(map_function, backend="cpu", map_t0=jax.pmap, map_D=jax.vmap)
vmap_gpu = partial(map_function, backend="gpu", map_t0=jax.vmap, map_D=jax.vmap)



from dataclasses import dataclass
from typing import List

import jax.numpy as jnp
import numpy as np
from scipy.linalg import block_diag
from tqdm.auto import tqdm

from nuance.nuance import Nuance
from nuance.search_data import SearchData


def solve_triangular(*gps_y):
    Ls = [gp.solver.solve_triangular(y) for gp, y in gps_y]
    if Ls[0].ndim == 1:
        return np.hstack(Ls)
    else:
        return block_diag(*Ls)


@dataclass
class CombinedNuance:
    """
    An object to combine `nuance` searches from multiple datasets.

    Parameters
    ----------
    datasets : List[Nuance]
        list of :py:class:`~nuance.Nuance` instances
    """

    datasets: List[Nuance]
    """Nuance instance of each dataset, where the linear search must be already ran."""
    search_data: SearchData = None
    """SearchData instance of the combined dataset."""

    def __post_init__(self):
        self._fill_search_data()
        self._compute_L()

    def __getitem__(self, i):
        return self.datasets[i]

    @property
    def model(self):
        """The model"""
        return self.datasets[0].model

    def _fill_search_data(self):
        if all([d.search_data is not None for d in self.datasets]):
            t0s = np.hstack([d.search_data.t0s for d in self.datasets])
            all_Ds = [n.search_data.Ds for n in self.datasets]
            all_equal = (
                np.diff(np.vstack(all_Ds).reshape(len(all_Ds), -1), axis=0) == 0
            ).all()
            assert (
                all_equal
            ), "All datasets linear searches must have same duration grids"
            Ds = self.datasets[0].search_data.Ds
            ll = np.vstack([d.search_data.ll for d in self.datasets])
            z = np.vstack([d.search_data.z for d in self.datasets])
            vz = np.vstack([d.search_data.vz for d in self.datasets])
            ll0 = np.sum([d.search_data.ll0 for d in self.datasets])

            self.search_data = SearchData(t0s, Ds, ll, z, vz, ll0)
        else:
            self.search_data = None

    @property
    def n(self):
        """Number of datasets"""
        return len(self.datasets)

    @property
    def time(self):
        """Time of all datasets"""
        return np.hstack([d.time for d in self.datasets])

    @property
    def flux(self):
        """Flux of all datasets"""
        return np.hstack([d.flux for d in self.datasets])

    def _compute_L(self):
        Liy = solve_triangular(*[(d.gp, d.flux) for d in self.datasets])
        LiX = solve_triangular(*[(d.gp, d.X.T) for d in self.datasets])

        def eval_model(ms):
            Lim = solve_triangular(*[(d.gp, m) for d, m in zip(self.datasets, ms)])
            LiXm = jnp.hstack([LiX, Lim[:, None]])
            LiXmT = LiXm.T
            LimX2 = LiXmT @ LiXm
            w = jnp.linalg.lstsq(LimX2, LiXmT @ Liy)[0]
            v = jnp.linalg.inv(LimX2)
            return w, v

        self.eval_model = eval_model

    def linear_search(
        self,
        t0s: np.ndarray,
        Ds: np.ndarray,
        positive: bool = True,
        progress: bool = True,
        backend: str = None,
        batch_size: int = None,
    ):
        """Performs the linear search for each dataset. Linear searches are saved as :py:class:`~nuance.SearchData`
        within each :py:class:`~nuance.Nuance` dataset.

        Parameters
        ----------
        t0s : np.ndarray
            array of model epochs
        Ds : np.ndarray
            array of model durations
        positive : bool, optional
            wether to force depth to be positive, by default True
        progress : bool, optional
            wether to show progress bar, by default True
        backend : str, optional
            backend to use, by default jax.default_backend() (options: "cpu", "gpu").
            This affects the linear search function jax-mapping strategy. For more details, see
            :py:func:`nuance.core.map_function`
        batch_size : int, optional
            batch size for parallel evaluation, by default None
        Returns
        -------
        None
        """
        for d in self.datasets:
            d.linear_search(
                t0s,
                Ds,
                progress=progress,
                backend=backend,
                batch_size=batch_size,
                positive=positive,
            )

    def solve(self, t0: float, D: float, P: float):
        """Solve the combined model for a given set of parameters.

        Parameters
        ----------
        t0 : float
            epoch, same unit as time
        D : float
            duration, same unit as time
        P : float, optionale
            period, same unit as time, by default None
        c : float, optional
            c parameter of the transit model, by default None

        Returns
        -------
        list
            (w, v): linear coefficients and their covariance matrix
        """
        models = [self.model(d.search_data.t0s, t0, D, P) for d in self.datasets]
        w, v = self.eval_model(models)
        return w, v

    def snr(self, t0: float, D: float, P: float):
        """SNR of transit linearly solved for epoch `t0` and duration `D` (and period `P` for a periodic transit)

        Parameters
        ----------
        t0 : float
            epoch, same unit as time
        D : float
            duration, same unit as time
        P : float, optional
            period, same unit as time, by default None
        c : float, optional
            c parameter of the transit model, by default None

        Returns
        -------
        float
            transit snr
        """
        w, v = self.solve(t0, D, P)
        return jnp.max(jnp.array([0, w[-1] / jnp.sqrt(v[-1, -1])]))

    def periodic_search(self, periods: np.ndarray, progress=True, dphi=0.01):
        """Performs the periodic search

        Parameters
        ----------
        periods : np.ndarray
            array of periods to search
        progress : bool, optional
            wether to show progress bar, by default True
        dphi: float, optional
            the relative step size of the phase grid. For each period, all likelihood quantities along time are
            interpolated along a phase grid of resolution `min(1/200, dphi/P))`. The smaller dphi
            the finer the grid, and the more resolved the transit epoch and period (the the more computationally expensive the
            periodic search). The default is 0.01.

        Returns
        -------
        :py:class:`nuance.SearchData`
            search results
        """
        n = len(periods)
        assert all(
            [d.search_data is not None for d in self.datasets]
        ), "linear search missing on at least one dataset"
        fold_functions = [d.search_data.fold_ll for d in self.datasets]
        Ds = self.search_data.Ds

        def _search(p):
            phase, _, p0 = fold_functions[0](p, dphi)
            P2 = p0 + np.sum([f(p)[2] for f in fold_functions[1:]], axis=0)
            i, j = np.unravel_index(np.argmax(P2), P2.shape)
            Ti = phase[i] * p
            return float(self.snr(Ti, Ds[j], p)), (
                Ti,
                Ds[j],
                p,
            )

        snr = np.zeros(n)
        params = np.zeros((n, 3))

        def _progress(x, **kwargs):
            return tqdm(x, **kwargs) if progress else x

        for i, p in enumerate(_progress(periods)):
            snr[i], params[i] = _search(p)

        new_search_data = self.search_data.copy()

        new_search_data.periods = periods
        new_search_data.Q_snr = snr
        new_search_data.Q_params = params

        return new_search_data

    def models(self, t0: float, D: float, P: float, split=False):
        """Solve the combined model for a given set of parameters.

        Parameters
        ----------
        t0 : float
            epoch, same unit as time
        D : float
            duration, same unit as time
        P : float, optional
            period, same unit as time, by default None

        Returns
        -------
        list
            (w, v): linear coefficients and their covariance matrix
        """
        ms = [d.model(d.time, t0, D, P) for d in self.datasets]
        w, _ = self.eval_model(ms)

        # means
        w_idxs = [0, *np.cumsum([d.X.shape[0] for d in self.datasets])]
        means = []
        for i in range(len(w_idxs) - 1):
            means.append(np.array(w[w_idxs[i] : w_idxs[i + 1]]) @ self.datasets[i].X)

        # signals
        signals = [self.model(d.time, t0, D, P=P) * w[-1] for d in self.datasets]

        # noises
        noises = []
        for i, d in enumerate(self.datasets):
            _, cond = d.gp.condition(d.flux - means[i] - signals[i])
            noises.append(cond.mean)
        if split:
            return means, signals, noises
        else:
            return np.hstack(means), np.hstack(signals), np.hstack(noises)

    def mask_model(self, t0: float, D: float, P: float):
        new_self = self.__class__(
            datasets=[d.mask_model(t0, D, P) for d in self.datasets], c=self.c
        )
        new_self._fill_search_data()
        new_self._compute_L()
        return new_self



import jax.numpy as jnp
from tinygp import GaussianProcess, kernels


def Rotation(sigma, period, Q0, dQ, f):
    """
    A kernel for a rotating star with a single mode of oscillation.
    """
    Q1 = 1 / 2 + Q0 + dQ
    w1 = (4 * jnp.pi * Q1) / (period * jnp.sqrt(4 * Q1**2 - 1))
    s1 = sigma**2 / ((1 + f) * w1 * Q1)

    Q2 = 1 / 2 + Q0
    w2 = (8 * jnp.pi * Q1) / (period * jnp.sqrt(4 * Q1**2 - 1))
    s2 = f * sigma**2 / ((1 + f) * w2 * Q2)
    kernel = kernels.quasisep.SHO(w1, Q1, s1) + kernels.quasisep.SHO(w2, Q2, s2)
    return kernel


def rotation(period=None, error=None, long_scale=0.5):
    initial_params = {
        "log_period": jnp.log(period) if period is not None else jnp.log(1.0),
        "log_Q": jnp.log(100),
        "log_sigma": jnp.log(1e-1),
        "log_dQ": jnp.log(100),
        "log_f": jnp.log(2.0),
        "log_short_scale": jnp.log(1e-2),
        "log_short_sigma": jnp.log(1e-2),
        "log_long_sigma": jnp.log(1e-2),
        "log_jitter": jnp.log(1.0) if error is None else jnp.log(error),
    }

    def build_gp(params, time):
        jitter2 = jnp.exp(2 * params["log_jitter"])
        short_scale = jnp.exp(params["log_short_scale"])
        short_sigma = jnp.exp(params["log_short_sigma"])
        long_sigma = jnp.exp(params["log_long_sigma"])

        kernel = (
            Rotation(
                jnp.exp(params["log_sigma"]),
                jnp.exp(params["log_period"]),
                jnp.exp(params["log_Q"]),
                jnp.exp(params["log_dQ"]),
                jnp.exp(params["log_f"]),
            )
            + kernels.quasisep.Exp(short_scale, short_sigma)
            + kernels.quasisep.Exp(long_scale, long_sigma)
        )

        return GaussianProcess(kernel, time, diag=jitter2, mean=0.0)

    return build_gp, initial_params



import jax

DEVICES_COUNT = jax.device_count()

config = jax.config
config.update("jax_enable_x64", True)

from nuance.combined import CombinedNuance
from nuance.nuance import Nuance
from nuance.search_data import SearchData
from nuance.star import Star
